import urllib
import sgmllib
import os
from jtalks.parser.PomFile import PomFile
from jtalks.util.Logger import Logger


class Nexus:
  """
  A class to work with Nexus (upload, download, search). Note, that for all the operations we need a build number
  to be passed into the constructor. Other properties may or may not be required (some of them can be determined,
  like project name and version by pom.xml).
  """
  base_url = "http://repo.jtalks.org/content/repositories/deployment-pipeline/"
  logger = Logger("Nexus")

  def __init__(self, build_number):
    self.build_number = build_number

  def upload_war(self, pom_file_location):
    """
      Uploads a war to the Nexus. The path to pom.xml is acceptaced as an argument, by this path
      we also determean where war file is placed.
    """
    pom = PomFile(pom_file_location)
    artifact_version = pom.version()
    artifact_id = pom.artifact_id()
    final_version = self.__final_version__(artifact_version)

    maven_deploy_command = ("mvn deploy:deploy-file -Durl={3} " +
                            "-DrepositoryId=deployment-pipeline -DgroupId=deployment-pipeline -DartifactId={0} -Dpackaging=war " +
                            "-Dfile={0}-view/{0}-web-view/target/{0}.war -Dversion={4}"
    ).format(artifact_id, artifact_version, self.build_number, self.base_url, final_version)
    print maven_deploy_command

    return_code = os.system(maven_deploy_command)
    if (return_code != 0):
      self.logger.error("Maven returned error code: " + str(return_code))
      raise Exception("Maven returned error code: " + str(return_code))

  def download_war(self, project):
    self.logger.info("Looking up build #{0} for {1} project", self.build_number, project)
    war_url = self.get_war_url(project, self.build_number)
    self.logger.info("Downloading artifact: [{0}]", war_url)
    urllib.urlretrieve(war_url, project + ".war")


  def get_war_url(self, project, build_number):
    group_id = "deployment-pipeline/"
    artifact_version_url = NexusPageWithVersions().parse(self.base_url + group_id + project).version(build_number)
    #get version by URL (last part is something like /jcommune/12.3.123/)
    artifact_version = artifact_version_url.rpartition(project + "/")[2].replace("/", "")
    return (artifact_version_url + "{0}-{1}.war".format(project, artifact_version))

  def __final_version__(self, artifact_version):
    """
     If build number is already in the version, then we should do nothing with it
    """
    if artifact_version.endswith("." + self.build_number):
      return artifact_version
    else:
      return artifact_version + "." + self.build_number


class NexusPageWithVersions(sgmllib.SGMLParser):
  hyperlinks = []

  def parse(self, url_with_versions):
    feed = urllib.urlopen(url_with_versions)
    page_content = feed.read()
    feed.close()

    self.feed(page_content)
    self.close()
    return self

  def start_a(self, attributes):
    for name, link in attributes:
      if name == "href":
        self.hyperlinks.append(link)

  def version(self, build_number):
    for link in self.hyperlinks:
      #project-x.y.BUILD_NUMBER - old format, project-x.y.BUILD_NUMBER.git_hash - new format
      if ".{0}/".format(build_number) in link or \
              ".{0}.".format(build_number) in link: return link
    raise Exception(
      "Couldn't find a build number {0}. Here are all the links: {1}".format(build_number, self.hyperlinks))



import xml.etree.ElementTree as ElementTree
from jtalks.parser.WrongConfigException import WrongConfigException
from jtalks.util.Logger import Logger

__author__ = 'stanislav bashkirtsev'


class TomcatServerXml:
  """
    Parses $TOMCAT_HOME/conf/server.xml file and returns attributes and tag values.
  """
  logger = Logger("TomcatServerXml")
  HTTP_PROTOCOL = "HTTP/1.1"
  tree = None

  def __init__(self, xml_element):
    """
      Creates a file from xml element object
      @param xml_element server.xml already parsed into xml.etree.Element
    """
    self.tree = xml_element

  @staticmethod
  def fromfile(filename):
    """
      Parses the specified server.xml
    """
    Logger("TomcatServerXml").info("Parsing [{0}] to obtain information about Tomcat", filename)
    return TomcatServerXml(ElementTree.parse(filename))

  @staticmethod
  def fromstring(file_content):
    """
      Creates object from string instead of from file
    """
    return TomcatServerXml(ElementTree.fromstring(file_content))

  def http_port(self):
    """
      Parses server.xml and obtains a Connector with protocol="HTTP/1.1". Returns its port.
    """
    http_connectors = self.tree.findall("./Service/Connector")
    http_connector = None
    for connector in http_connectors:
      if connector.get("protocol") == self.HTTP_PROTOCOL:
        http_connector = connector
        break
    if http_connector is None:
      error_message = "$TOMCAT_HOME/conf/server.xml didn't contain Connector with HTTP/1.1 protocol"
      self.logger.error(error_message)
      raise WrongConfigException(error_message)
    return http_connector.get("port")


from jtalks.util.Logger import Logger

__author__ = 'stanislav bashkirtsev'


class DeployToTomcatFacade:
  NEXUS_URL = "http://repo.jtalks.org/content/repositories/deployment-pipeline/deployment-pipeline/"
  logger = Logger("DeployToTomcatFacade")

  def __init__(self, application_context):
    self.app_context = application_context

  def deploy(self):
    script_settings = self.app_context.script_settings
    self.__raise_if_settings_not_specified__(script_settings)

    if script_settings.grab_envs == "true":
      self.app_context.environment_config_grabber().grab_jtalks_configs()
    self.app_context.nexus().download_war(project=self.app_context.script_settings.project)
    self.app_context.tomcat().deploy_war()
    self.app_context.sanity_test().check_app_started_correctly()

  def __raise_if_settings_not_specified__(self, script_settings):
    if script_settings.build is None:
      self.logger.error("Build number was not specified, see [{0}] to get list of builds", self.NEXUS_URL)
      raise RuntimeError
    if script_settings.project not in ["jcommune", "poulpe", "antarcticle"]:
      self.logger.error("A correct project should be specified: [poulpe, jcommune, antarcticle]")
      raise RuntimeError



import os
import MySQLdb

from jtalks.util.Logger import Logger


class DbOperations:
  """
    Base class for working with database
  """
  logger = Logger("DbBase")
  env = None
  db_settings = None

  def __init__(self, env, db_settings):
    """
    Args:
    env: environment name e.g. dev, uat
    db_settings: database connection settings of type DbSettings, which will be used for all
                 database related operations
    """
    self.env = env
    self.db_settings = db_settings


  def connect_to_database(self):
    """
    Connects to the specified database
    """
    self.logger.info("Connecting to [{0}] with user [{1}]", self.db_settings.dbHost, self.db_settings.dbUser)
    self.connection = MySQLdb.connect(host=self.db_settings.dbHost, user=self.db_settings.dbUser,
                                      passwd=self.db_settings.dbPass)
    self.cursor = self.connection.cursor()

  def close_connection(self):
    """
    Closes open connection to the database
    """
    self.connection.close()

  def recreate_database(self):
    """
    Deletes the specified database and create it again
    """
    self.logger.info("Dropping and creating database from scratch: [{0}]", self.db_settings.dbName)
    self.cursor.execute('DROP DATABASE IF EXISTS ' + self.db_settings.dbName)
    self.cursor.execute('CREATE DATABASE ' + self.db_settings.dbName)

  def backup_database(self, backupPath):
    """
    Backups database to given backupPath
    """
    self.logger.info("Backing up database [{0}] to [{1}] using user [{2}]",
                     self.db_settings.dbName, backupPath, self.db_settings.dbUser)
    os.popen("mysqldump -u{0} -p{1} {2} > {3}/{2}.sql"
      .format(self.db_settings.dbUser, self.db_settings.dbPass, self.db_settings.dbName, backupPath)) \
      .read()
    self.logger.info("Database backed up: environment=[{0}]".format(self.env))

  def restore_database_from_file(self, backupPath):
    """
    Restores database from file specified in backupPath
    """
    self.logger.info("Loading a db dump from [{0}/{1}.sql]", backupPath, self.db_settings.dbName)
    self.recreate_database()
    os.popen('mysql -u{0} -p{1} {2} < {3}/{2}.sql'
      .format(self.db_settings.dbUser,self.db_settings.dbPass, self.db_settings.dbName, backupPath))\
      .read()
    self.logger.info("Database restored: environment=[{0}]".format(self.env))






import xml.etree.ElementTree as ElementTree


class PomFile:
  """
    Represents Maven POM file and provides information inside of it. For instance we might need to get
    a version of the artifact: we need to pass a file to the class and get version:
    PomFile("pom.xml").version()

    @author stanislav bashkirtsev
  """

  # Maven XML schema, we need it to search for tags that obey this schema.
  maven_schema = "{http://maven.apache.org/POM/4.0.0}"
  tree = None

  def __init__(self, file_name):
    self.tree = ElementTree.parse(file(file_name))

  def version(self):
    """
      The version of the artifact inside of pom.xml. Note, that some modules can inherit version
      from its parents, thus version might be absent and None will be returned in this case.
    """
    version = self.get_tag_text_by_name("version")
    return version.replace("-SNAPSHOT", "")

  def artifact_id(self):
    """
      Gets the artifactId from pom.xml file. If file is not valid and there is no such tag, then None will be returned.
    """
    return self.get_tag_text_by_name("artifactId")

  def get_tag_text_by_name(self, tag_name):
    """
      Parses pom.xml file and finds the tag with specified name in it. It takes into account Maven XML schema as well.
    """
    return self.tree.getroot().findtext("{0}{1}".format(self.maven_schema, tag_name))



import time

import requests

from jtalks.sanity.SanityCheckFailedException import SanityCheckFailedException
from jtalks.util.Logger import Logger


__author__ = 'stanislav bashkirtsev'


class SanityTest:
  """
    Tests that application was deployed correctly without errors. For these purposes it opens some pages and determines
    whether they return HTML. If let's say they return HTTP 500, then the test failed. It has to break CI builds.
  """
  HOST = "127.0.0.1"
  SLEEP_TIME = 10
  logger = Logger("SanityTest")
  port = None
  app_name = None

  def __init__(self, tomcat_port, app_name, sanity_test_timeout_sec=120):
    """
     @param tomcat_port - an HTTP port to access the web server where application is
    """
    self.port = int(tomcat_port)
    self.app_name = app_name
    self.sanity_test_timeout_sec = sanity_test_timeout_sec
    if app_name == "ROOT":
      self.app_name = ""

  def check_app_started_correctly(self):
    self.logger.info("Sleeping for {0} seconds to give Tomcat time to start", self.SLEEP_TIME)
    time.sleep(self.SLEEP_TIME)
    request_address = "http://{0}:{1}/{2}".format(self.HOST, self.port, self.app_name)
    self.logger.info("Running sanity tests to check whether application started correctly and responds back..")
    self.logger.info("Connecting to {0}", request_address)
    response = requests.get(request_address, timeout=self.sanity_test_timeout_sec)
    if response.status_code not in [200, 201]:
      self.logger.error("While accessing main page, app answered with error: [{0} {1} {2}]",
                        response.status_code, response.reason, response.text)
      raise SanityCheckFailedException("Sanity check failed")
    self.logger.info("Sanity check passed: [{0} {1}]", response.status_code, response.reason)



from ConfigParser import ConfigParser, NoSectionError
import os
from jtalks.util.Logger import Logger

__author__ = 'stanislav bashkirtsev'


class ScriptSettings:
  SCRIPT_TEMD_DIR = '/tmp/jtalks-cicd/'
  script_work_dir = "" + os.path.expanduser("~/.jtalks/")
  backups_dir = script_work_dir + "backups/"
  ENV_CONFIGS_DIR = script_work_dir + "environments/"
  GLOBAL_CONFIG_LOCATION = ENV_CONFIGS_DIR + "global-configuration.cfg"
  logger = Logger("ScriptSettings")

  def __init__(self, build, project=None, env=None, grab_envs=None, work_dir=None, sanity_test_timeout_sec=120,
               package_version=None):
    """
     @param grab_envs - whether or not we should clone JTalks predefined environment configuration from private git
     repo
     @param work_dir - standard is ~/.jtalks, but it may be useful to override this value, e.g. during tests
     @param sanity_test_timeout_sec - how much time do sanity tests wait for the application to respond until they
            consider deployment as failed
    """
    self.env = env
    self.build = build
    self.project = project
    self.grab_envs = grab_envs
    self.script_work_dir = work_dir
    self.sanity_test_timeout_sec = sanity_test_timeout_sec
    self.package_version = package_version

  def log_settings(self):
    self.logger.info(
      "Script Settings: project=[{0}], env=[{1}], build number=[{2}], sanity test timeout=[{3}], package version=[{4}]",
      self.project, self.env, self.build, self.sanity_test_timeout_sec, self.package_version)
    self.logger.info("Environment configuration: [{0}]", self.ENV_CONFIGS_DIR)

  def create_work_dir_if_absent(self):
    self.__create_dir_if_absent__(self.script_work_dir)
    self.__create_dir_if_absent__(self.get_env_configs_dir())
    self.__create_dir_if_absent__(self.get_backup_folder())

  def __create_dir_if_absent__(self, directory):
    if not os.path.exists(directory):
      self.logger.info("Creating directory [{0}]", directory)
      os.mkdir(directory)

  def get_tomcat_location(self):
    """
      Gets value of the tomcat home from [project].cfg file related to particular env and project
    """
    return self.__get_property('tomcat', 'location')

  def get_app_final_name(self):
    """
      Gets the name of the application to be deployed (even if it's Poulpe, it can be deployed as ROOT.war).
    """
    return self.__get_property('app', 'final_name')

  def get_temp_dir(self):
    """
     Script can save there some temp files.
    """
    return self.SCRIPT_TEMD_DIR

  def get_tomcat_port(self):
    """
      This is not actually a pre-configured parameter, it parses $TOMCAT_HOME/conf/server.xml to find out the HTTP
      port it's going to be listening. This is needed e.g. for sanity tests to.
    """
    server_xml = os.path.join(self.get_tomcat_location(), "conf", "server.xml")
    raise RuntimeError("tomcat port parsing is not implemented yet")

  def get_backup_folder(self):
    return self.backups_dir

  def get_env_configs_dir(self):
    return self.ENV_CONFIGS_DIR

  def get_global_config_location(self):
    return self.GLOBAL_CONFIG_LOCATION

  def __get_property(self, section, prop_name):
    """
      Finds property first in project configuration, then environment configuration and if there is no such property
      there, then it looks is it up in global configuration.
      @param section - a section joins several properties under it
      @param prop_name - a particular property name from specified section
      @returns None if there is no such property found in any config
    """
    value = self.__get_project_property(section, prop_name)
    if value == None:
      value = self.__get_env_property(section, prop_name)
    if value == None:
      value = self.__get_global_prop(section, prop_name)
    if value == None:
      self.logger.error("Property [{0}] was not found in any configs", prop_name)
      raise ValueError
    return self.__replace_placeholders(value)

  def __get_project_property(self, section, prop_name):
    """
      Finds property value in project configuration. This overrides env and global configuration.
    """
    config = ConfigParser()
    config.read(os.path.join(self.ENV_CONFIGS_DIR, self.env, self.project + ".cfg"))
    return self.__get_value_from_config(config, section, prop_name)

  def __get_env_property(self, section, prop_name):
    """
      Finds property value in configs/${env}/${env}.cfg configuration file. This overrides global configuration, but
      still can be overriden by project configs. These configs are shared between apps of the same environment. E.g.
      if we have Poulpe and JCommune on UAT env, and there is a file configs/uat/uat.cfg, then these properties are
      shared between those Poulpe and JCommune.
    """
    config = ConfigParser()
    config.read(os.path.join(self.ENV_CONFIGS_DIR, self.env, "environment-configuration.cfg"))
    return self.__get_value_from_config(config, section, prop_name)

  def __get_global_prop(self, section, prop_name):
    """
      Finds a property value in configs/global-configuration.cfg.
    """
    config = ConfigParser()
    config.read(self.GLOBAL_CONFIG_LOCATION)
    return self.__get_value_from_config(config, section, prop_name)

  def __get_value_from_config(self, config, section, prop_name):
    try:
      return config.get(section, prop_name)
    except NoSectionError:
      return None

  def __replace_placeholders(self, prop_value):
    """
      Replaces placeholder for env and project that were possibly set in config files.
    """
    return prop_value.replace("${env}", self.env).replace("${project}", self.project)

  def get_sanity_test_timeout_sec(self):
    """
      How much time do sanity tests wait for the application response until they consider that the app didn't start
    """
    self.sanity_test_timeout_sec


#!/usr/bin/env python
from distutils.core import setup
from setuptools import find_packages
from jtalks import __version__

__author__ = 'stanislav bashkirtsev'

setup(name='jtalks-cicd',
      version=__version__,
      description='Installs JTalks apps like jcommune and poulpe configuring them. Usually used by CI to implement CD (continuous delivery)',
      author='Stanislav Bashkirtsev',
      author_email='stanislav.bashkirtsev@gmail.com',
      url='http://github.com/jtalks-org/jtalks-cicd',
      install_requires=['distribute>=0.6.35', 'requests', 'mock', 'GitPython', 'paramiko', 'mysql-connector-python',
                        'MySQL-python'],
      scripts=['bin/jtalks'],
      packages=find_packages(),
      # after package installation we'll have a nice directory instead of zipped artifact which is
      # easier to work with
      zip_safe=False
      #include_package_data=True should add MANIFEST.in first to include non-py files if needed
)



from datetime import datetime


class Logger:
  className = None

  def __init__(self, className):
    self.className = className

  def info(self, message, *args):
    self.__log__(colors.OKGREEN, "INFO", message, args)

  def warn(self, message, *args):
    self.__log__(colors.WARNING, "WARN", message, args)

  def error(self, message, *args):
    self.__log__(colors.FAIL, "ERROR", message, args)

  def __log__(self, color, level, message, args):
    print(color + "[{0}][JTALKS][{1}][{2}] ".format(datetime.now().isoformat(), level, self.className) +
          message.format(*args) + colors.END)


class colors:
  HEADER = '\033[35m'
  OKBLUE = '\033[34m'
  OKGREEN = '\033[32m'
  WARNING = '\033[33m'
  FAIL = '\033[31m'
  END = '\033[0m'

  def disable(self):
    self.HEADER = ''
    self.OKBLUE = ''
    self.OKGREEN = ''
    self.WARNING = ''
    self.FAIL = ''
    self.ENDC = ''



from datetime import datetime
import os
import shutil

from jtalks.util.Logger import Logger


class Backuper:
  """
    Responsible for backing up artifacts like war files, config files, as well as DB backups.
    Stores backups in the folder and allows to restore those backups from it.
    @author stanislav bashkirtsev
  """
  logger = Logger("Backuper")
  BACKUP_FOLDER_DATE_FORMAT = "%Y_%m_%dT%H_%M_%S"

  def __init__(self, backup_folder, script_settings, db_operations):
    """
      @param backup_folder a directory to put backups to, it may contain backups from different envs,
        thus a folder for each env will be created
      @param script_settings is instance of ScriptSettings class
      @param db_operations instance of DbOperations
    """
    if not backup_folder.endswith("/"):
      raise ValueError("Folder name should finish with '/'. Actual value: " + backup_folder)
    self.root_backup_folder = backup_folder
    self.script_settings = script_settings
    self.db_operations = db_operations

  def backup(self):
    """
    Does all operations for project backup:
     - backup the application database
     - backup application war-file
     - backup tomcat and ehcache configuration files
    """
    folder_to_put_backups = self.create_folder_to_backup(self.root_backup_folder, self.script_settings)
    self.backup_tomcat(folder_to_put_backups)
    self.backup_db(folder_to_put_backups)

  def clean_old_backups(self, backups_to_keep):
    """
    In order not to keep a lot of useless backups that take space, we're purging old ones.
    @param backups_to_keep amount of backups to keep in backup folder without removal
    """
    all_backups = self.get_list_of_backups()

  def create_folder_to_backup(self, backup_folder, script_settings):
    """
      We don't just put backups into a backup folder, we're creating a new folder there
      with current date so that our previous backups are kept there. E.g.:
      '/var/backups/prod/20130302T195924'

      @param backups_folder - basic folder to put backups there, it will be concatenated with env and current time
      @param script_settings to figure out what's the env and what's the project we're going to backup
    """
    now = datetime.now().strftime(self.BACKUP_FOLDER_DATE_FORMAT)

    final_backup_folder = "{0}/{1}".format(self.get_project_backup_folder(), now)
    os.makedirs(final_backup_folder)
    self.logger.info("Backing up old resources to [{0}]", final_backup_folder)
    return final_backup_folder

  def backup_tomcat(self, backup_folder):
    """
      To be safe that we didn't forget anything, we're doing a backup for the whole tomcat directory.
      @param backup_folder a directory to put Tomcat to, should be created by the time this method is invoked
    """
    tomcat_location = self.script_settings.get_tomcat_location()
    if os.path.exists(tomcat_location):
      shutil.copytree(tomcat_location, backup_folder + "/tomcat")
    else:
      self.logger.info("There was no previous tomcat folder [{0}], so nothing to backup", tomcat_location)

  def backup_db(self, backup_folder):
    self.db_operations.backup_database(backup_folder)

  def get_list_of_backups(self):
    """
      Gets the list of backups saved for
    """
    return os.listdir(self.root_backup_folder)

  def get_project_backup_folder(self):
    return "{0}{1}/{2}".format(self.root_backup_folder, self.script_settings.env, self.script_settings.project)




__author__ = 'stanislav bashkirtsev'


class LoadDbFromBackup:
  def __init__(self, db, ssh):
    self.db = db
    self.ssh = ssh

  def load(self):
    self.ssh.download_backup_prod_db()
    self.db.connect_to_database()
    self.db.restore_database_from_file(self.ssh.sftpBackupFileName)
    self.db.update_properties_to_preprod()
    self.db.connection.close()
    self.ssh.remove_backup_prod_db_file()



import os
import ConfigParser

import paramiko
from jtalks.util.Logger import Logger


class SSH:
  env = None
  config = None
  sftpTransport = None
  sftpClient = None
  sftpHost = None
  sftpPort = None
  sftpUser = None
  sftpPass = None
  sftpBackupArchive = None
  sftpBackupFileName = None
  logger = Logger("SSH")

  def __init__(self, script_settings):
    self.env = script_settings.env
    self.config = ConfigParser.ConfigParser()
    self.config.read(script_settings.get_env_configs_dir() + self.env + "/ssh.cfg")
    self.sftpHost = self.config.get('sftp', 'sftp_host')
    self.sftpPort = self.config.get('sftp', 'sftp_port')
    self.sftpUser = self.config.get('sftp', 'sftp_user')
    self.sftpPass = self.config.get('sftp', 'sftp_pass')
    self.sftpBackupArchive = self.config.get('sftp', 'sftp_backup_archive')
    self.sftpBackupFileName = self.config.get('sftp', 'sftp_backup_filename')

  def download_backup_prod_db(self):
    self.sftp_connection()
    attrList = self.sftpClient.listdir_attr('.')
    attrTimeTmp = 0
    dirWithLastUpdate = None
    for attr in attrList:
      if attrTimeTmp < attr.st_mtime:
        attrTimeTmp = attr.st_mtime
        dirWithLastUpdate = attr.filename
    self.sftpClient.get(dirWithLastUpdate + '/' + self.sftpBackupArchive, './' + self.sftpBackupArchive)
    os.popen('bunzip2 -d ' + self.sftpBackupArchive)
    self.sftp_connection_close()

  def sftp_connection_close(self):
    self.sftpClient.close()
    self.sftpTransport.close()

  def sftp_connection(self):
    self.logger.info("Getting a DB backup from {0}", self.sftpHost)
    self.sftpTransport = paramiko.Transport((self.sftpHost, int(self.sftpPort)))
    self.sftpTransport.connect(username=self.sftpUser, password=self.sftpPass)
    self.sftpClient = paramiko.SFTPClient.from_transport(self.sftpTransport)

  def remove_backup_prod_db_file(self):
    os.popen('rm -rf ' + self.sftpBackupFileName)  

  
